//////////////////////////////////////////////////////////////////////
// Bind.cpp
// -------------------------------------------------------------------
// The bind module used to bind keystrokes to commands.
//
// <thohell@home.se>
//////////////////////////////////////////////////////////////////////
#include ".\\External\\ClientCore.cpp"

// Functions as they appear in the source...
BOOL PRIVATE OnGameCommandSet(char** argv, int argc);
BOOL PRIVATE OnGameCommandAdd(char** argv, int argc);
BOOL PRIVATE OnGameCommandRotate(char** argv, int argc);
BOOL PRIVATE OnGameCommandUnset(char** argv, int argc);
BOOL PRIVATE OnGameCommandKeycode(char** argv, int argc);
BOOL PRIVATE OnGameCommandList(char** argv, int argc);
BOOL PRIVATE OnGameCommandLoad(char** argv, int argc);
BOOL PRIVATE OnGameCommandSave(char** argv, int argc);
BOOL PRIVATE OnGameCommandClear(char** argv, int argc);
void GameKeyDownSTUB(); //0x6FACEF99
DWORD __fastcall GameKeyDown(DWORD aKeyCode);

// Some global stuff
FINGERPRINTSTRUCT fps;
BOOL	ShowKeycode=FALSE;
char	CurrentSection[64];

typedef struct command_list_t
{
	char*	Command;
	command_list_t*	Next_List;
} COMMAND_LIST;

typedef struct keys_t
{
	BOOL	Rotational;
	DWORD	Returncode;
	command_list_t*	List;
} KEYS;

KEYS k[256];
BOOL Loaded=FALSE;

//////////////////////////////////////////////////////////////////////
// CLIENTINFO
// -------------------------------------------------------------------
// You *need* this, don't ask why :)
//
// 0.01 2001-10-14 First version
// 0.02 2001-10-14 Added game-hook for keystrokes
// 0.03 2001-10-14 Commands commands commands...
// 0.04 2001-10-14 Fixed maphack compatibility problem
// 0.05 2001-10-14 Added Returncode to re-map existing keys
// 0.06 2001-10-14 Dont remember :)
// 1.00 2001-11-21 Small changes for D2Loader compatibility, matured
// 2.00 2001-12-09 ClckWcher - Multiple Commands Can now be binded to keys;Added "Add" Command
// 2.01 2001-12-09 ClckWcher - Speed Increase;Added Rotational Components
// 2.02 2001-12-25 ClckWcher - Add | translator
// 2.03 2002-02-13 bootyjuice - "Fixed" up so it works with D2Hackit 0.56+
// 2.04 2002-08-25 Abin - Fixed a wild-pointer bug in "OnGameCommandUnset" and "OnGameCommandSet"
//////////////////////////////////////////////////////////////////////
DWORD	ModuleVersion=MAKELONG(2,3);
char	ModuleAuthor[]="thohell ; ClckWcher ; bootyjuice ; Abin";
char	ModuleWebsite[]="people.atl.attbroadband.com/bootyjuice";
char	ModuleDescription[]="Bind module - Binds keystrokes to commands";
char	ModuleEmail[]="thohell@home.se ; ClckWcher@aol.com ; b00tyju1ce@yahoo.com ; abinn32@yahoo.com";

//////////////////////////////////////////////////////////////////////
// MODULECOMMANDSTRUCT ModuleCommands[]
// -------------------------------------------------------------------
// To add your own commands, just add them to this list like:
//
//	{
//		"name"			// Name of the command
//		pFunction		// The function that handles command
//		"usage"			// Help text for the command
//	}
// 
// Functions should use following convention...
// BOOL PRIVATE OnGameCommandName(char** argv, int argc)
// ...and return TRUE on success or FALSE to show usage of the command.
//
// To use your commands in the game, just type 
// .modulename <command> [arguments]
//////////////////////////////////////////////////////////////////////

MODULECOMMANDSTRUCT ModuleCommands[]=
{
	{	// The 'help' command is implemented in ClientCore.cpp and 
		// should not need to be changed.
		"help",
		OnGameCommandHelp,
		"helpc0 List commands available in this module.\n"
		"<command> helpc0 Shows detailed help for <command> in this module."
	},
	{
		"set",
		OnGameCommandSet,
		"set <keycode> <returncode> [command1[;command2;...;commandN]]c0 Binds the key with <keycode> to return <returncode> and perform [command..commandN]."
	},
	{
		"add",
		OnGameCommandAdd,
		"add <keycode> [command1[;command2;...;commandN]]c0 Binds the key with <keycode> to return <returncode> and perform [command1..commandN]."
	},
	{
		"rotate",
		OnGameCommandRotate,
		"rotate <keycode>c0 Toggles the weither or not commands rotate."
	},
	{
		"unset",
		OnGameCommandUnset,
		"unset <keycode>c0 Removes the binding of key with <keycode>."
	},
	{
		"keycode",
		OnGameCommandKeycode,
		"keycodec0 Shows the keycode of the next key you press."
	},
	{
		"list",
		OnGameCommandList,
		"listc0 Lists all currently assigned keys."
	},
	{	"load",
		OnGameCommandLoad,
		"load [section]c0 Loads key-bindings in [section] from bind.ini."
	},
	{	"save",
		OnGameCommandSave,
		"save [section]c0 Saves current key-bindings to <section> in bind.ini."
	},
	{	"clear",
		OnGameCommandClear,
		"clearc0 Removes all key-bindings."
	},
	{NULL}	// No more commands
};

//////////////////////////////////////////////////////////////////////
// OnClientStart
// -------------------------------------------------------------------
// Runs *once* before client is loaded. Return FALSE to prevent from
// loading.
//////////////////////////////////////////////////////////////////////
BOOL EXPORT OnClientStart()
{
	// Initiate list
	char *argv[3];
	char argc=3;
	argv[0]="";
	argv[1]="load";
	argv[2]="default";
	OnGameCommandLoad(argv, argc);
		
	// Find location to intercept.
	if(!server->GetFingerprint("D2HackIt", "GameKeyDownIntercept", fps))
		return FALSE;


	sprintf(CurrentSection, "Default");

	// Patch code
	server->Intercept(INST_CALL, fps.AddressFound, (DWORD)&GameKeyDownSTUB, fps.PatchSize);
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientStop
// -------------------------------------------------------------------
// Runs *once* before client is unloaded. Return value should be TRUE,
// but it is not used at this time.
//////////////////////////////////////////////////////////////////////
BOOL EXPORT OnClientStop()
{
	// Unpatch code
	server->Intercept(INST_CALL, (DWORD)&GameKeyDownSTUB, fps.AddressFound, fps.PatchSize);

	// Save
	char *argv[2];//ClckWcher
	char argc=2;
	argv[0]="";
	argv[1]="save";
	OnGameCommandSave(argv, argc);
	COMMAND_LIST * List, * Next_List;

	// Empty array
	for (int i=0; i!=256; i++)
	if (k[i].List)
	{
		List = k[i].List , Next_List;
		while(List != NULL)
		{
			Next_List = List->Next_List;
			delete List->Command;
			delete List;
			List = Next_List;
		}
	}

	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandSet
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandSet(char** argv, int argc)
{
	// If command line is shorter than 4 tokens, something is wrong
	if (argc<4) return FALSE;
	
	char t[1024];	// Should be enough...
	COMMAND_LIST *List, *Next_List;
	char *ptSeperator;

	char*x="";
	DWORD Keycode=(strtoul(argv[2], &x, 0x10)&0xff);
	if (strlen(x))
	{
		sprintf(t, "'%s' is not a valid keycode!", argv[2]);
		server->GamePrintError(t);
		return TRUE;
	}

	DWORD Returncode=(strtoul(argv[3], &x, 0x10)&0xff);
	if (strlen(x))
	{
		sprintf(t, "'%s' is not a valid returncode!", argv[2]);
		server->GamePrintError(t);
		return TRUE;
	}

	k[Keycode].Returncode = Returncode;
	k[Keycode].Rotational = FALSE;

	if (k[Keycode].List)
	{
		List = k[Keycode].List , Next_List;

		// Added by Abin(8/25/2002):
		// Wild-pointer removal:
		// Crashing may occur without below line:
		k[Keycode].List = NULL;

		while(List != NULL)
		{
			Next_List = List->Next_List;
			delete List->Command;
			delete List;
			List = Next_List;
		}
	}

	if (argc>4)
	{
		t[0]=0;
		List = k[Keycode].List = new COMMAND_LIST;
		for (int i=4; i!=argc; i++)
		{
			if((ptSeperator = strchr(argv[i],'`')) != NULL)
			{
				*ptSeperator = '|';
				i --;
			}
			else if((ptSeperator = strchr(argv[i],';')) != NULL)
			{ 
				if(ptSeperator != argv[i])
				{
					*ptSeperator = NULL;
					strcat(t, argv[i]);
					strcat(t, " ");
					*ptSeperator = ';';
				}
				strcpy(List->Command=new char[strlen(t)+1],t);
				List = List->Next_List = new COMMAND_LIST;
				t[0] = 0;
				argv[i] = ptSeperator;
				argv[i--] ++;
			}
			else if(* argv[i] != NULL)
			{
				strcat(t, argv[i]);
				strcat(t, " ");
			}
		}
		List->Next_List = NULL;
		strcpy(List->Command=new char[strlen(t)+1],t);
	} 

	sprintf(t, "Binding %.2x -> %.2x ", Keycode, Returncode);
	if (k[Keycode].List)
	{
		strcat(t, "Action: ");
		List = k[Keycode].List;
		while(List->Next_List != NULL)
		{
			strcat(t, List->Command);
			strcat(t, " ; ");
			List = List->Next_List;
		}
		strcat(t, List->Command);
	}
	server->GamePrintVerbose(t);
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandAdd
// - ClckWcher
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandAdd(char** argv, int argc)
{
	// If command line is shorter than 4 tokens, something is wrong
	if (argc<4) return FALSE;
	
	char t[1024];	// Should be enough...
	COMMAND_LIST *List;
	char *ptSeperator;

	char*x="";
	DWORD Keycode=(strtoul(argv[2], &x, 0x10)&0xff);
	if (strlen(x))
	{
		sprintf(t, "'%s' is not a valid keycode!", argv[2]);
		server->GamePrintError(t);
		return TRUE;
	}
	
	t[0]=0;
	if((List = k[Keycode].List) != NULL)
	{
		while(List->Next_List != NULL)
		{
			List = List->Next_List;
		}
	}
	else
	{
		sprintf(t, "'%s' is not an active keycode!", argv[2]);
		server->GamePrintError(t);
		return TRUE;
	}
	List = List->Next_List = new COMMAND_LIST;

	for (int i=3; i!=argc; i++)
	{
		if((ptSeperator = strchr(argv[i],'`')) != NULL)
		{
			*ptSeperator = '|';
			i --;
		}
		if((ptSeperator = strchr(argv[i],';')) != NULL)
		{ 
			if(ptSeperator != argv[i])
			{
				*ptSeperator = NULL;
				strcat(t, argv[i]);
				strcat(t, " ");
			}
			strcpy(List->Command = new char[strlen(t)+1] , t );
			List = List->Next_List = new COMMAND_LIST;
			t[0] = 0;
			argv[i] = ptSeperator;
			argv[i--] ++;
		}
		else if(* argv[i] != NULL)
		{
			strcat(t, argv[i]);
			strcat(t, " ");
		}
	}
	List->Next_List = NULL;
	strcpy(List->Command = new char[strlen(t)+1] ,t);
	
	sprintf(t, "Adding to %.2x", Keycode);
	if (k[Keycode].List)
	{
		strcat(t, " Action: ");
		List = k[Keycode].List;
		while(List->Next_List != NULL)
		{
			strcat(t, List->Command);
			strcat(t, ";");
			List = List->Next_List;
		}
		strcat(t, List->Command);
	}
	server->GamePrintVerbose(t);
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandRotate
// - ClckWcher
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandRotate(char** argv, int argc)
{
	// If command line does not equal 3 tokens than something is wrong
	if (argc==2) return FALSE;
	
	char t[1024];	// Should be enough...
	COMMAND_LIST *List;
	char *x = "";

	DWORD Keycode=(strtoul(argv[2], &x, 0x10)&0xff);
	if (strlen(x))
	{
		sprintf(t, "'%s' is not a valid keycode!", argv[2]);
		server->GamePrintError(t);
		return TRUE;
	}
	
	t[0]=0;
	if((List = k[Keycode].List) != NULL)
	{
		if(List->Next_List == NULL)
		{
			sprintf(t, "'%s' you have to have more than one command to rotate!", argv[2]);
			server->GamePrintError(t);
			return TRUE;
		}
		if(k[Keycode].Rotational)
		{
			k[Keycode].Rotational = FALSE;
			sprintf(t, "%.2X Rotate off",Keycode);
		}
		else
		{
			k[Keycode].Rotational = TRUE;
			sprintf(t, "%.2X Rotate on",Keycode);
		}
	}
	else
	{
		sprintf(t, "'%s' is not an active keycode!", argv[2]);
		server->GamePrintError(t);
		return TRUE;
	}
	server->GamePrintVerbose(t);
	return TRUE;
}
//////////////////////////////////////////////////////////////////////
// OnClientCommandUnset
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandUnset(char** argv, int argc)
{
	// If command line is not 3 tokens, something is wrong
	if (argc!=3) return FALSE;

	char t[128];
	COMMAND_LIST *List, *Next_List;

	// Check for valid keycode
	char*x="";
	DWORD Keycode=(strtoul(argv[2], &x, 0x10)&0xff);
	if (strlen(x))
	{
		sprintf(t, "'%s' is not a valid keycode!", argv[2]);
		server->GamePrintError(t);
		return TRUE;
	}

	// Restore returncode
	k[Keycode].Returncode=Keycode;

	// Remove command, if any
	if (k[Keycode].List)
	{
		List = k[Keycode].List , Next_List;

		// Added by Abin(8/25/2002):
		// Crashing may occur without below line:
		k[Keycode].List = NULL;

		while(List != NULL)
		{
			Next_List = List->Next_List;
			delete List->Command;
			delete List;
			List = Next_List;
		}
	}

	sprintf(t, "Keycode '%.2x' restored.", Keycode);
	server->GamePrintVerbose(t);

	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandKeydcode
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandKeycode(char** argv, int argc)
{
	// If command line is not 2 tokens, something is wrong
	if (argc!=2) return FALSE;

	server->GamePrintInfo("Press a key to get its keycode...");
	ShowKeycode=TRUE;
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandList
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandList(char** argv, int argc)
{
	// If command line is not 2 tokens, something is wrong
	if (argc!=2) return FALSE;
	
	server->GamePrintInfo("Keys currently mapped by bind:");
	// Loop through struct
	for (int i=0; i!=256; i++)
	{
		// Found mapped key
		if ((k[i].Returncode != i) || (k[i].List))
		{
			char t[1024];
			COMMAND_LIST* List;
			sprintf(t, "%.2x -> %.2x", i, k[i].Returncode);
			if(k[i].Rotational)
				strcat(t, " Rotate");
			if (k[i].List)
			{
				strcat(t, " Action: ");
				List = k[i].List;
				while(List->Next_List != NULL)
				{
					strcat(t, List->Command);
					strcat(t, " ; ");
					List = List->Next_List;
				}
				strcat(t, List->Command);
			}
			server->GamePrintInfo(t);
		}
	}
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandLoad
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandLoad(char** argv, int argc)
{
	// If command line is not 2 or 3 tokens, something is wrong
	if ((argc!=3) && (argc!=2))return FALSE;

	char t[1024];
	char *retval;
	COMMAND_LIST *List, *Next_List;

	if (argc == 3)
		strcpy(CurrentSection, argv[2]);

	sprintf(t, "Loading key-mappings from section: c4[%s]",CurrentSection);
	server->GamePrintInfo(t);
	
	for (int i=0; i!=256; i++)
	{
		k[i].Returncode = i;
		k[i].Rotational = FALSE;
		if (k[i].List)
		{
			List = k[i].List , Next_List;
			k[i].List = NULL;
			while(List != NULL)
			{
				Next_List = List->Next_List;
				delete List->Command;
				delete List;
				List = Next_List;
			}
		}
		sprintf(t, "Keycode%.2x", i);
		retval=server->GetHackProfileString("bind", CurrentSection, t);
		
		// If we found something, process it
		if (strlen(retval))
		{
			char *x="";
			char Returncode[3];
			memcpy(Returncode,retval,2);
			Returncode[2] = NULL;
			// Save returncode
			if (strlen(retval)>3)
			{
				if(retval[2] == '0')
					k[i].Rotational = FALSE;
				else
					k[i].Rotational = TRUE;
			}
			k[i].Returncode=(strtoul(Returncode, &x, 0x10)&0xff);
			// If there is a command-string, save it
			char *ptSeperator;
			char *Command = &retval[4];
			if (strlen(retval)>4)
			{
				List = k[i].List = new COMMAND_LIST;
				while((ptSeperator = strchr(Command,';')) != NULL)
				{
					* ptSeperator = 0;
					strcpy(List->Command = new char[strlen(Command) + 1],Command);
					List = List->Next_List = new COMMAND_LIST;
					Command = ptSeperator;
					Command ++;
				}
				
				List->Next_List = NULL;
				strcpy(List->Command = new char[strlen(Command) + 1],Command);
			}
			if ((k[i].Returncode != i) || (k[i].List != NULL))
			{
				sprintf(t, "%.2X->%.2X", i, k[i].Returncode);
				if(k[i].Rotational)
					strcat(t, " Rotate");
				if(List = k[i].List)
				{
					strcat(t, " Action ");
					while(List->Next_List != NULL)
					{
						strcat(t, List->Command );
						strcat(t, " ; " );
						List = List->Next_List;
					}
					strcat(t, List->Command);
				}
				server->GamePrintVerbose(t);
			}
		}
		// delete retval;
	}
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandSave
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandSave(char** argv, int argc)
{
	// If command line is not 2 or 3 tokens, something is wrong
	if ((argc!=3) && (argc!=2))return FALSE;

	char t[1024];
	char key[12];
	COMMAND_LIST* List;

	if (argc == 3)
		strcpy(CurrentSection, argv[2]);

	sprintf(t, "Saving key-mappings to section: c4[%s]",CurrentSection);
	server->GamePrintInfo(t);

	// Loop through struct
	for (int i=0; i!=256; i++)
	{
		// Found mapped key
		sprintf(key, "Keycode%.2X", i);
		sprintf(t, "%.2X", k[i].Returncode);
		if(k[i].Rotational)
			strcat(t, "1");
		else
			strcat(t, "0");
		
		List = k[i].List;
		while(List != NULL)
		{
			strcat(t, ";");
			strcat(t, List->Command);
			List = List->Next_List;
		}
		server->SetHackProfileString("bind", CurrentSection, key, t);
	}
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// OnClientCommandClear
// -------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnGameCommandClear(char** argv, int argc)
{
	// If command line is not 2 tokens, something is wrong
	if (argc!=2) return FALSE;
	COMMAND_LIST *List, *Next_List;
	
	for (int i=0; i!=256; i++)
	{
		k[i].Returncode=i;
		List = k[i].List;
		k[i].List = NULL;
		k[i].Rotational = FALSE;
		while(List != NULL)
		{
			Next_List = List->Next_List;
			delete List->Command;
			delete List;
			List = Next_List;
		}
	}

	server->GamePrintInfo("All key-bindings has been removed.");

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// GameKeyDownSTUB()
// --------------------------------------------------------------------------
// esi = Keycode
/////////////////////////////////////////////////////////////////////////////
void __declspec(naked) GameKeyDownSTUB()
{
	__asm {
		nop								// Make room for original code
		nop
		nop
		nop
		nop
		nop
		pushad
		mov ecx, esi
		call GameKeyDown
		// Shuffle... This is ugly!
		push eax						// Push return code
		add esp, 4						// Move stack
		popad							// Restore registers
		sub esp, 0x24					// move stack
		pop esi							// Pop return code
		add esp, 0x20					// Move stack
		ret
	}
}

/////////////////////////////////////////////////////////////////////////////
// GameKeyDown()
// --------------------------------------------------------------------------
// Our 'clean' GameKeyDown() 
/////////////////////////////////////////////////////////////////////////////
DWORD __fastcall GameKeyDown(DWORD aKeyCode)
{
	COMMAND_LIST *List, *Next_List;
	if (ShowKeycode)
	{
		ShowKeycode=FALSE;
		char t[16];
		sprintf(t, "Keycode: %.2x", aKeyCode&0xff);
		server->GamePrintInfo(t);
		return 0;
	}
	
	if(k[aKeyCode&0xff].List)
	{
		List = k[aKeyCode&0xff].List;
		if(k[aKeyCode&0xff].Rotational)
		{
			if(List->Next_List != NULL)
			{
				Next_List = k[aKeyCode&0xff].List = List->Next_List;

				while(Next_List->Next_List != NULL)
					Next_List = Next_List->Next_List;
				Next_List->Next_List = List;
				List->Next_List = NULL;
			}
			server->GameCommandLine(List->Command);
		}
		else
			while(List != NULL)
			{
				server->GameCommandLine(List->Command);
				List = List->Next_List;
			}
	}

	return k[aKeyCode&0xff].Returncode;
}


